from .tensor import *
from . import functional_cv2 as F_cv2
from . import functional_pil as F_pil
from . import functional_tensor as F_tensor
import cv2 as cv


def im_type(img):
    if is_numpy(img):
        return 'cv2'
    if is_pil(img):
        return 'pil'
    if is_tensor(img):
        return 'tensor'

def call_by_name(func, img, *args, **kwargs):
    func_map = {
        'cv2': F_cv2,
        'pil': F_pil,
        'tensor': F_tensor
    }
    return getattr(func_map[im_type(img)], func)(img, *args, **kwargs)
    

def inter_code(mode, backend):
    interpolation = {
        'pil': {
            'nearest': Image.NEAREST,
            'bilinear': Image.BILINEAR,
            'bicubic': Image.BICUBIC,
        },
        'cv2': {
            'nearest': cv.INTER_NEAREST,
            'bilinear': cv.INTER_LINEAR,
            'bicubic': cv.INTER_CUBIC,
        },
        'tensor': {},
    }
    return interpolation[backend].get(mode, mode)


def get_image_size(img):
    if is_pil(img):
        w, h = img.size
    elif is_tensor(img) and img.dim() > 2:
        h, w = img.shape[-2:]
    elif is_numpy(img):
        h, w = img.shape[:2]
    else:
        raise TypeError("Unexpected type {}".format(type(img)))
    return h, w


def hflip(img):
    return call_by_name('hflip', img)


def vflip(img):
    return call_by_name('vflip', img)


def rotate(img, degree, interpolation='bilinear', expand=False, center=None):
    return call_by_name('rotate', img, degree=degree, interpolation=inter_code(interpolation, im_type(img)), expand=False, center=None)


def resize(img, size, interpolation):
    return call_by_name('resize', img, size=size, interpolation=inter_code(interpolation, im_type(img)))

def crop(img, i, j, h, w):
    return call_by_name('crop', img, i=i, j=j, h=h, w=w)

def resized_crop(img, i, j, h, w, size, interpolation):
    return call_by_name('resized_crop', img, i, j, h, w, size, inter_code(interpolation, im_type(img)))

def pad(img, padding, fill,  padding_mode):
    return call_by_name('pad', img, padding, fill, padding_mode)

def adjust_brightness(img, scale):
    return call_by_name('adjust_brightness', img, scale)

def adjust_contrast(img, scale):
    return call_by_name('adjust_contrast', img, scale)

def adjust_saturation(img, scale):
    return call_by_name('adjust_saturation', img, scale)

def adjust_hue(img, scale):
    return call_by_name('adjust_hue', img, scale)

def adjust_gamma(img, gamma, gain):
    return call_by_name('adjust_gamma', img, gamma, gain)

def perspective(img, startpoints, endpoints, interpolation):
    return call_by_name('perspective', startpoints, endpoints, inter_code(interpolation, im_type(img)))

def label_map(img, table):
    return call_by_name('label_map', img, table)

def normalize(tensor, mean, std):
    return F_tensor.normalize(tensor, mean, std)




import cv2 as cv
import matplotlib.pyplot as plt
import numpy as np
import torch

import cvt

img = cv.imread('assets/image.jpg')
mask = cv.imread('assets/mask.png')
data = {'image':img, 'mask':mask}

tfms = cvt.from_file('configs/transform1.json')
data = tfms(data)
plt.imshow(data['image'])
plt.show()
plt.imshow(data['mask'])
plt.show()


img = cv.imread('assets/image.jpg')
mask = cv.imread('assets/mask.png')
data = {'image':img, 'mask':mask}

tfms = cvt.from_file('configs/transform2.json')
data = tfms(data)
print('data:', data['image'].shape, type(data['image']))
print('mask:', data['mask'].shape, type(data['mask']))
print('mask:', data['mask'].unique())


from cvt.transform import Resize, RandomCrop, Sequence

tfms = Sequence([
    Resize([256, 256]),
    RandomCrop([224, 224])
])

import cv2 as cv 
img = cv.imread('test/assets/image.jpg')[:,:,[2,1,0]]
data = {'image':img}
data = tfms(data)
import matplotlib.pyplot as plt
plt.imshow(data['image'])
plt.show()


import torch 
import numpy as np 
from PIL import Image

def is_numpy(img):
    return isinstance(img, np.ndarray)

def is_pil(img):
    return isinstance(img, Image.Image)

def is_tensor(img):
    return isinstance(img, torch.Tensor)

def to_numpy(img):
    if is_tensor(img):
        return img.detach().cpu().numpy()
    return np.array(img)

def to_pil(img):
    if is_tensor(img):
        img = to_numpy(img)
    if is_numpy(img):
        return Image.fromarray(np.uint8(img))
    return img

def to_tensor(pic):

    if isinstance(pic, np.ndarray):
        # handle numpy array
        if pic.ndim == 2:
            pic = pic[:, :, None]

        img = torch.from_numpy(pic.transpose((2, 0, 1)))
        # backward compatibility
        if isinstance(img, torch.ByteTensor):
            return img.float().div(255)
        else:
            return img

    # handle PIL Image
    if pic.mode == 'I':
        img = torch.from_numpy(np.array(pic, np.int32, copy=False))
    elif pic.mode == 'I;16':
        img = torch.from_numpy(np.array(pic, np.int16, copy=False))
    elif pic.mode == 'F':
        img = torch.from_numpy(np.array(pic, np.float32, copy=False))
    elif pic.mode == '1':
        img = 255 * torch.from_numpy(np.array(pic, np.uint8, copy=False))
    else:
        img = torch.ByteTensor(torch.ByteStorage.from_buffer(pic.tobytes()))
    # PIL image mode: L, LA, P, I, F, RGB, YCbCr, RGBA, CMYK
    if pic.mode == 'YCbCr':
        nchannel = 3
    elif pic.mode == 'I;16':
        nchannel = 1
    else:
        nchannel = len(pic.mode)
    img = img.view(pic.size[1], pic.size[0], nchannel)
    img = img.transpose(0, 1).transpose(0, 2).contiguous()
    if isinstance(img, torch.ByteTensor):
        return img.float().div(255)
    else:
        return img



from cvt.transform import (RandomBright, RandomContrast, RandomHorizontalFlip,
                           RandomHue, RandomRotation, RandomVerticalFlip,
                           Sample, Sequence, Shuffle)

tfms = Sequence([
    Sample([RandomBright(0.2), RandomContrast(0.2), RandomHue(0.2)]),
    Shuffle([RandomHorizontalFlip(), RandomVerticalFlip(), RandomRotation(30)])
])

import cv2 as cv 
img = cv.imread('test/assets/image.jpg')[:,:,[2,1,0]]
data = {'image':img}
data = tfms(data)
import matplotlib.pyplot as plt
plt.imshow(data['image'])
plt.show()


from PIL import Image, ImageOps, ImageEnhance
import numpy as np

def hflip(img):
    return img.transpose(Image.FLIP_LEFT_RIGHT)


def vflip(img):
    return img.transpose(Image.FLIP_TOP_BOTTOM)


def crop(img, i, j, h, w):
    return img.crop((j, i, j + w, i + h))



def pad(img, padding, fill=0, padding_mode='constant'):
    r"""Pad the given PIL Image on all sides with specified padding mode and fill value.

    Args:
        img (PIL Image): Image to be padded.
        padding (int or tuple): Padding on each border. If a single int is provided this
            is used to pad all borders. If tuple of length 2 is provided this is the padding
            on left/right and top/bottom respectively. If a tuple of length 4 is provided
            this is the padding for the left, top, right and bottom borders
            respectively.
        fill: Pixel fill value for constant fill. Default is 0. If a tuple of
            length 3, it is used to fill R, G, B channels respectively.
            This value is only used when the padding_mode is constant
        padding_mode: Type of padding. Should be: constant, edge, reflect or symmetric. Default is constant.

            - constant: pads with a constant value, this value is specified with fill

            - edge: pads with the last value on the edge of the image

            - reflect: pads with reflection of image (without repeating the last value on the edge)

                       padding [1, 2, 3, 4] with 2 elements on both sides in reflect mode
                       will result in [3, 2, 1, 2, 3, 4, 3, 2]

            - symmetric: pads with reflection of image (repeating the last value on the edge)

                         padding [1, 2, 3, 4] with 2 elements on both sides in symmetric mode
                         will result in [2, 1, 1, 2, 3, 4, 4, 3]

    Returns:
        PIL Image: Padded image.
    """

    if padding_mode == 'constant':
        if img.mode == 'P':
            palette = img.getpalette()
            image = ImageOps.expand(img, border=padding, fill=fill)
            image.putpalette(palette)
            return image

        return ImageOps.expand(img, border=padding, fill=fill)
    else:
        if isinstance(padding, int):
            pad_left = pad_right = pad_top = pad_bottom = padding
        if isinstance(padding, Sequence) and len(padding) == 2:
            pad_left = pad_right = padding[0]
            pad_top = pad_bottom = padding[1]
        if isinstance(padding, Sequence) and len(padding) == 4:
            pad_left = padding[0]
            pad_top = padding[1]
            pad_right = padding[2]
            pad_bottom = padding[3]

        if img.mode == 'P':
            palette = img.getpalette()
            img = np.asarray(img)
            img = np.pad(img, ((pad_top, pad_bottom), (pad_left, pad_right)), padding_mode)
            img = Image.fromarray(img)
            img.putpalette(palette)
            return img

        img = np.asarray(img)
        # RGB image
        if len(img.shape) == 3:
            img = np.pad(img, ((pad_top, pad_bottom), (pad_left, pad_right), (0, 0)), padding_mode)
        # Grayscale image
        if len(img.shape) == 2:
            img = np.pad(img, ((pad_top, pad_bottom), (pad_left, pad_right)), padding_mode)

        return Image.fromarray(img)



def resize(img, size, interpolation=Image.BILINEAR):
    if isinstance(size, int):
        w, h = img.size
        if (w <= h and w == size) or (h <= w and h == size):
            return img
        if w < h:
            ow = size
            oh = int(size * h / w)
            return img.resize((ow, oh), interpolation)
        else:
            oh = size
            ow = int(size * w / h)
    else:
        oh, ow = size
    return img.resize((ow, oh), interpolation)


def rotate(img, degree, interpolation=Image.BILINEAR, expand=False, center=None, fill=0):
    return img.rotate(degree, interpolation, expand, center, fillcolor=fill)

def resized_crop(img, i, j, h, w, size, interpolation=Image.BILINEAR):
    img = crop(img, i, j, h, w)
    img = resize(img, size, interpolation)
    return img

def adjust_brightness(img, scale):
    enhancer = ImageEnhance.Brightness(img)
    img = enhancer.enhance(scale)
    return img

def adjust_contrast(img, scale):
    enhancer = ImageEnhance.Contrast(img)
    img = enhancer.enhance(scale)
    return img

def adjust_saturation(img, scale):
    enhancer = ImageEnhance.Color(img)
    img = enhancer.enhance(scale)
    return img


def adjust_hue(img, scale):

    if not(-0.5 <= scale <= 0.5):
        raise ValueError('hue_factor is not in [-0.5, 0.5].'.format(scale))

    input_mode = img.mode
    if input_mode in {'L', '1', 'I', 'F'}:
        return img

    h, s, v = img.convert('HSV').split()

    np_h = np.array(h, dtype=np.uint8)
    # uint8 addition take cares of rotation across boundaries
    with np.errstate(over='ignore'):
        np_h += np.uint8(scale * 255)
    h = Image.fromarray(np_h, 'L')

    img = Image.merge('HSV', (h, s, v)).convert(input_mode)
    return img


def adjust_gamma(img, gamma, gain=1):
    input_mode = img.mode
    img = img.convert('RGB')

    gamma_map = [255 * gain * pow(ele / 255., gamma) for ele in range(256)] * 3
    img = img.point(gamma_map) 
    img = img.convert(input_mode)
    return img

def _get_perspective_coeffs(startpoints, endpoints):
    matrix = []

    for p1, p2 in zip(endpoints, startpoints):
        matrix.append([p1[0], p1[1], 1, 0, 0, 0, -p2[0] * p1[0], -p2[0] * p1[1]])
        matrix.append([0, 0, 0, p1[0], p1[1], 1, -p2[1] * p1[0], -p2[1] * p1[1]])

    A = torch.tensor(matrix, dtype=torch.float)
    B = torch.tensor(startpoints, dtype=torch.float).view(8)
    res = torch.lstsq(B, A)[0]
    return res.squeeze_(1).tolist()


def perspective(img, startpoints, endpoints, interpolation=Image.BICUBIC):
    coeffs = _get_perspective_coeffs(startpoints, endpoints)
    return img.transform(img.size, Image.PERSPECTIVE, coeffs, interpolation)

def label_map(img, tabel, value=0):
    img = img.point(lambda x: tabel.get(x, value))
    return img




import collections
import json
import os

from .transform import (Identity, LabelMap, Normalize, Pad, RandomBright,
                        RandomContrast, RandomCrop, RandomGamma,
                        RandomHorizontalFlip, RandomHue, RandomResizedCrop,
                        RandomRotation, RandomSaturation, RandomVerticalFlip,
                        Resize, Sample, Sequence, Shuffle, ToTensor)

tfms_map = {
    'identity': Identity,
    
    'bright': RandomBright,
    'contrast': RandomContrast,
    'gamma': RandomGamma,
    'saturation': RandomSaturation,
    'hue': RandomHue,

    'hflip': RandomHorizontalFlip,
    'vflip': RandomVerticalFlip,
    'resized_crop': RandomResizedCrop,
    'resize': Resize,
    'rotate': RandomRotation,
    'crop': RandomCrop,
    'pad': Pad,

    'sequence': Sequence,
    'shuffle':  Shuffle,
    'sample': Sample,

    'totensor':  ToTensor,
    'label_map': LabelMap,
    'normalize': Normalize,
}


def _get_tfms_from_dict(cfg):
    tfms = []
    for tfm, params in cfg.items():
        assert tfm in tfms_map
        if tfm in ['sequence', 'shuffle', 'sample']:
            t = _get_tfms_from_dict(params)
            tfms.append(tfms_map[tfm](t))
            continue
        tfms.append(tfms_map[tfm](**params))
    if len(tfms) == 1 and isinstance(tfms[0], Sequence):
        return tfms[0]
    return tfms

def from_dict(cfg):
    if len(cfg) == 1 and list(cfg.keys())[0] == 'sequence':
        pass
    else:
        cfg = collections.OrderedDict([('sequence', cfg)])
    return _get_tfms_from_dict(cfg)

def from_file(filename):
    if filename.endswith('json'):
        cfg_dict = json.load(
            open(filename), object_pairs_hook=collections.OrderedDict)
        return from_dict(cfg_dict)


def from_option():
    pass


def from_yaml():
    pass



import json
import math
import numbers
import random
from abc import ABCMeta, abstractmethod
from collections import Iterable, OrderedDict
from functools import partial
from typing import List

import numpy as np
import torch

from . import functional as F


class Sequence:
    def __init__(self, transforms):
        if not isinstance(transforms, List):
            transforms = [transforms]
        self.transforms = transforms

    def __call__(self, data):
        for t in self.transforms:
            data = t(data)
        return data
    
    def __repr__(self):
        format_string = self.__class__.__name__ + '('
        for t in self.transforms:
            format_string += '\n'
            format_string += '    {0}'.format(t)
        format_string += '\n)'
        return format_string

    def to_dict(self):
        return OrderedDict([("sequence", OrderedDict([t.to_dict().popitem() for t in self.transforms]))])

    def to_json(self, indent=4):
        return json.dumps(self.to_dict(), indent=indent)

class Shuffle:
    def __init__(self, transforms):
        if not isinstance(transforms, List):
            transforms = [transforms]
        self.transforms = transforms

    def __call__(self, data):
        random.shuffle(self.transforms)
        for t in self.transforms:
            data = t(data)
        return data

    def __repr__(self):
        format_string = self.__class__.__name__ + '('
        for t in self.transforms:
            format_string += '\n'
            format_string += '    {0}'.format(t)
        format_string += '\n)'
        return format_string

    def to_dict(self):
        return OrderedDict([("shuffle", OrderedDict([t.to_dict().popitem() for t in self.transforms]))])

    def to_json(self, indent=4):
        return json.dumps(self.to_dict(), indent=indent)

class Sample:
    def __init__(self, transforms, k=1):
        if not isinstance(transforms, List):
            transforms = [transforms]
        assert len(transforms) >= k
        self.transforms = transforms
        self.k = k

    def __call__(self, data):
        transforms = random.sample(self.transforms, k=self.k)
        for t in transforms:
            data = t(data)
        return data

    def __repr__(self):
        format_string = self.__class__.__name__ + '('
        for t in self.transforms:
            format_string += '\n'
            format_string += '    {0}'.format(t)
        format_string += '\n)'
        return format_string

    def to_dict(self):
        return OrderedDict([("sample", OrderedDict([t.to_dict().popitem() for t in self.transforms]))])

    def to_json(self, indent=4):
        return json.dumps(self.to_dict(), indent=indent)

class Transform(metaclass=ABCMeta):
    @abstractmethod
    def apply_image(self, img):
        pass

    def apply_mask(self, mask):
        raise NotImplementedError

    def apply_box(self, box):
        raise NotImplementedError

    def apply_kp(self, kp):
        raise NotImplementedError

    def __repr__(self):
        format_string = self.__class__.__name__
        return format_string

    @abstractmethod
    def to_dict(self):
        pass

class ToTensor(Transform):
    def __call__(self, data):
        if isinstance(data, dict):
            for k, v in data.items():
                if v is not None:
                    data[k] = getattr(self, f'apply_{k}')(v)
        else:
            data = self.apply_image(data)
        return data

    def apply_image(self, img):
        return F.to_tensor(img)

    def apply_mask(self, mask):
        if isinstance(mask, List):
            return torch.stack([torch.as_tensor(np.asarray(m), dtype=torch.int64) for m in mask])
        return torch.as_tensor(np.asarray(mask), dtype=torch.int64)

    def to_dict(self):
        return OrderedDict([("totensor", {})])

class Normalize(Transform):
    def __init__(self, mean, std):
        self.mean = mean
        self.std = std

    def __call__(self, data):
        if isinstance(data, dict):
            for k, v in data.items():
                if v is not None:
                    data[k] = getattr(self, f'apply_{k}')(v)
        else:
            data = self.apply_image(data)
        return data
    
    def apply_image(self, img):
        return F.normalize(img, self.mean, self.std)

    def apply_mask(self, mask):
        return mask

    def to_dict(self):
        return OrderedDict([("normalize", {"mean": list(self.mean), "std": list(self.std)})])

class LabelMap(Transform):
    def __init__(self, table):
        table = {int(k):int(v) for k,v in table.items()}
        self.label_map = table

    def __call__(self, data):
        if isinstance(data, dict):
            for k, v in data.items():
                if v is not None:
                    data[k] = getattr(self, f'apply_{k}')(v)
        else:
            data = self.apply_image(data)
        return data
    
    def apply_image(self, img):
        return img

    def apply_mask(self, mask):
        if isinstance(mask, List):
            return mask
        return F.label_map(mask, self.label_map)

    def to_dict(self):
        return OrderedDict([("labem_map", {"table": self.label_map})])

class Identity(Transform):
    def __call__(self, data):
        return data

    def apply_image(self, img):
        return img

    def to_dict(self):
        return OrderedDict([("identity", {})])

class RandomHorizontalFlip(Transform):
    def __init__(self, p=0.5):
        self.p = p

    def __call__(self, data):
        if random.random() < self.p:
            if isinstance(data, dict):
                for k, v in data.items():
                    if v is not None:
                        data[k] = getattr(self, f'apply_{k}')(v)
            else:
                data = self.apply_image(data)
        return data

    def apply_image(self, img):
        return F.hflip(img)

    def apply_mask(self, mask):
        if isinstance(mask, List):
            return [F.hflip(m) for m in mask]
        return F.hflip(mask)

    def to_dict(self):
        return OrderedDict([('hflip', {'p': self.p})])


class RandomVerticalFlip(Transform):
    def __init__(self, p=0.5):
        self.p = p

    def __call__(self, data):
        if random.random() < self.p:
            if isinstance(data, dict):
                for k, v in data.items():
                    if v is not None:
                        data[k] = getattr(self, f'apply_{k}')(v)
            else:
                data = self.apply_image(data)
        return data

    def apply_image(self, img):
        return F.vflip(img)

    def apply_mask(self, mask):
        if isinstance(mask, List):
            return [F.vflip(m) for m in mask]
        return F.vflip(mask)

    def to_dict(self):
        return OrderedDict([('vflip', {'p': self.p})])


class RandomRotation(Transform):
    def __init__(self, degrees, interpolation='nearest', expand=False, center=None, fill=0):
        if isinstance(degrees, numbers.Number):
            if degrees < 0:
                raise ValueError(
                    "If degrees is a single number, it must be positive.")
            self.degrees = (-degrees, degrees)
        else:
            if len(degrees) != 2:
                raise ValueError(
                    "If degrees is a sequence, it must be of len 2.")
            self.degrees = degrees

        self.interpolation = interpolation
        self.expand = expand
        self.center = center
        self.fill = fill

    def __call__(self, data):
        degree = self.get_params(self.degrees)
        if isinstance(data, dict):
            for k, v in data.items():
                if v is not None:
                    data[k] = getattr(self, f'apply_{k}')(v, degree)
        else:
            data = self.apply_image(data, degree)
        return data

    @staticmethod
    def get_params(degrees):
        degree = random.uniform(degrees[0], degrees[1])
        return degree

    def apply_image(self, img, degree):
        return F.rotate(img, degree, self.interpolation, self.expand, self.center)

    def apply_mask(self, mask, degree):
        rotate = partial(F.rotate, degree=degree, interpolation='nearest',
                         expand=self.expand, center=self.center)
        if isinstance(mask, List):
            return [rotate(m) for m in mask]
        return rotate(mask)

    def to_dict(self):
        return OrderedDict([('rotate', {
            'degrees': self.degrees,
            'interpolation': self.interpolation,
            'expand': self.expand,
            'center': self.center,
            'fill': self.fill
        })])


class Resize(Transform):
    def __init__(self, size, interpolation='bilinear'):
        assert isinstance(size, int) or (
            isinstance(size, Iterable) and len(size) == 2)
        self.size = size
        self.interpolation = interpolation

    def __call__(self, data):
        if isinstance(data, dict):
            for k, v in data.items():
                if v is not None:
                    data[k] = getattr(self, f'apply_{k}')(v)
        else:
            data = self.apply_image(data)
        return data

    def apply_image(self, img):
        return F.resize(img, self.size, self.interpolation)

    def apply_mask(self, mask):
        resize = partial(F.resize, size=self.size, interpolation='nearest')
        if isinstance(mask, List):
            return [resize(m) for m in mask]
        return resize(mask)

    def to_dict(self):
        return OrderedDict([('resize', {'size': self.size, 'interpolation': self.interpolation})])


class RandomCrop(Transform):
    def __init__(self, size):
        if isinstance(size, numbers.Number):
            self.size = (int(size), int(size))
        else:
            self.size = size
    
    @staticmethod
    def get_params(img, output_size):
        """Get parameters for ``crop`` for a random crop.

        Args:
            img (PIL Image): Image to be cropped.
            output_size (tuple): Expected output size of the crop.

        Returns:
            tuple: params (i, j, h, w) to be passed to ``crop`` for random crop.
        """
        w, h = F.get_image_size(img)
        th, tw = output_size
        if w == tw and h == th:
            return 0, 0, h, w

        i = random.randint(0, h - th)
        j = random.randint(0, w - tw)
        return i, j, th, tw
    
    def __call__(self, data):
        if isinstance(data, dict):
            params = self.get_params(data['image'], self.size)
            for k, v in data.items():
                if v is not None:
                    data[k] = getattr(self, f'apply_{k}')(v, *params)
        else:
            params = self.get_params(data, self.size)
            data = self.apply_image(data, *params)
        return data

    def apply_image(self, img, i, j, h, w):
        return F.crop(img, i, j, h, w)

    def  apply_mask(self, mask, i, j, h, w):
        if isinstance(mask, List):
            return [F.crop(m, i, j, h, w) for m in mask]
        return F.crop(mask, i, j, h, w)

    def to_dict(self):
        return OrderedDict([('crop', {'size': self.size})])

class RandomResizedCrop(Transform):
    def __init__(self, size, scale=(0.08, 1.0), ratio=(3. / 4., 4. / 3.), interpolation='bilinear'):
        self.size = size
        self.scale = scale
        self.ratio = ratio
        self.interpolation = interpolation

    def __call__(self, data):
        if isinstance(data, dict):
            params = self.get_params(data['image'], self.scale, self.ratio)
            for k, v in data.items():
                if v is not None:
                    data[k] = getattr(self, f'apply_{k}')(v, *params)
        else:
            params = self.get_params(data, self.scale, self.ratio)
            data = self.apply_image(data, *params)
        return data

    @staticmethod
    def get_params(img, scale, ratio):

        width, height = F.get_image_size(img)
        area = height * width

        for attempt in range(10):
            target_area = random.uniform(*scale) * area
            log_ratio = (math.log(ratio[0]), math.log(ratio[1]))
            aspect_ratio = math.exp(random.uniform(*log_ratio))

            w = int(round(math.sqrt(target_area * aspect_ratio)))
            h = int(round(math.sqrt(target_area / aspect_ratio)))

            if 0 < w <= width and 0 < h <= height:
                i = random.randint(0, height - h)
                j = random.randint(0, width - w)
                return i, j, h, w

        # Fallback to central crop
        in_ratio = float(width) / float(height)
        if (in_ratio < min(ratio)):
            w = width
            h = int(round(w / min(ratio)))
        elif (in_ratio > max(ratio)):
            h = height
            w = int(round(h * max(ratio)))
        else:  # whole image
            w = width
            h = height
        i = (height - h) // 2
        j = (width - w) // 2
        return i, j, h, w

    def apply_image(self, img, i, j, h, w):
        return F.resized_crop(img, i, j, h, w, self.size, self.interpolation)

    def apply_mask(self, mask, i, j, h, w):
        if isinstance(mask, List):
            return [F.resized_crop(m, i, j, h, w, self.size, 'nearest') for m in mask]
        return F.resized_crop(mask, i, j, h, w, self.size, 'nearest')

    def to_dict(self):
        return OrderedDict([('resized_crop', {
            'size': self.size,
            'scale': self.scale,
            'ratio': self.ratio,
            'interpolation': self.interpolation
        })])


class Pad(Transform):
    def __init__(self, padding, fill=0, padding_mode='constant'):
        assert isinstance(padding, (numbers.Number, tuple))
        assert isinstance(fill, (numbers.Number, str, tuple))
        assert padding_mode in ['constant', 'edge', 'reflect', 'symmetric']
        if isinstance(padding, Sequence) and len(padding) not in [2, 4]:
            raise ValueError("Padding must be an int or a 2, or 4 element tuple, not a " +
                             "{} element tuple".format(len(padding)))

        self.padding = padding
        self.fill = fill
        self.padding_mode = padding_mode

    def __call__(self, data):
        if isinstance(data, dict):
            for k, v in data.items():
                if v is not None:
                    data[k] = getattr(self, f'apply_{k}')(v, self.padding, self.fill, self.padding_mode)
        else:
            data = self.apply_image(data, self.padding, self.fill, self.padding_mode)
        return data

    def apply_image(self, img, padding, fill, padding_mode):
        return F.pad(img, padding, fill, padding_mode)

    def apply_mask(self, mask, padding, fill, padding_mode):
        if isinstance(mask, List):
            return [F.pad(m, padding, fill, padding_mode) for m in mask]
        return F.pad(mask, padding, fill, padding_mode)
    
    def to_dict(self):
        return OrderedDict([('pad', {'padding': self.padding, 'fill':self.fill,  'padding_mode':self.padding_mode})])


class RandomBright(Transform):
    def __init__(self, scale):
        if isinstance(scale, numbers.Number):
            scale = (1-scale, 1+scale)
        self.scale = scale

    def __call__(self, data):
        scale = random.uniform(*self.scale)
        if isinstance(data, dict):
            for k, v in data.items():
                if v is not None:
                    data[k] = getattr(self, f'apply_{k}')(v, scale)
        else:
            data = self.apply_image(data, scale)
        return data

    def apply_image(self, img, scale):
        return F.adjust_brightness(img, scale)

    def apply_mask(self, mask, scale):
        return mask

    def to_dict(self):
        return OrderedDict([('bright', {'scale': self.scale})])

class RandomContrast(Transform):
    def __init__(self, scale):
        if isinstance(scale, numbers.Number):
            scale = (1-scale, 1+scale)
        self.scale = scale

    def __call__(self, data):
        scale = random.uniform(*self.scale)
        if isinstance(data, dict):
            for k, v in data.items():
                if v is not None:
                    data[k] = getattr(self, f'apply_{k}')(v, scale)
        else:
            data = self.apply_image(data, scale)
        return data

    def apply_image(self, img, scale):
        return F.adjust_contrast(img, scale)

    def apply_mask(self, mask, scale):
        return mask

    def to_dict(self):
        return OrderedDict([('contrast', {'scale': self.scale})])


class RandomSaturation(Transform):
    def __init__(self, scale):
        if isinstance(scale, numbers.Number):
            scale = (1-scale, 1+scale)
        self.scale = scale

    def __call__(self, data):
        scale = random.uniform(*self.scale)
        if isinstance(data, dict):
            for k, v in data.items():
                if v is not None:
                    data[k] = getattr(self, f'apply_{k}')(v, scale)
        else:
            data = self.apply_image(data, scale)
        return data

    def apply_image(self, img, scale):
        return F.adjust_saturation(img, scale)

    def apply_mask(self, mask, scale):
        return mask

    def to_dict(self):
        return OrderedDict([('saturation', {'scale': self.scale})])


class RandomHue(Transform):
    def __init__(self, scale):
        if isinstance(scale, numbers.Number):
            assert abs(scale) <= 0.5
            scale = (-scale, scale)
        self.scale = scale

    def __call__(self, data):
        scale = random.uniform(*self.scale)
        if isinstance(data, dict):
            for k, v in data.items():
                if v is not None:
                    data[k] = getattr(self, f'apply_{k}')(v, scale)
        else:
            data = self.apply_image(data, scale)
        return data

    def apply_image(self, img, scale):
        return F.adjust_hue(img, scale)

    def apply_mask(self, mask, scale):
        return mask

    def to_dict(self):
        return OrderedDict([('hue', {'scale': self.scale})])


class RandomGamma(Transform):
    def __init__(self, scale, gain=1):
        if isinstance(scale, numbers.Number):
            scale = (1-scale, 1+scale)
        self.scale = scale
        self.gain = gain

    def __call__(self, data):
        scale = random.uniform(*self.scale)
        if isinstance(data, dict):
            for k, v in data.items():
                if v is not None:
                    data[k] = getattr(self, f'apply_{k}')(v, scale)
        else:
            data = self.apply_image(data, scale)
        return data

    def apply_image(self, img, scale):
        return F.adjust_gamma(img, scale, self.gain)

    def apply_mask(self, mask, scale):
        return mask

    def to_dict(self):
        return OrderedDict([('gamma', {'scale': self.scale})])


class RandomPerspective(object):
    def __init__(self, distortion_scale=0.5, p=0.5, interpolation='bicubic'):
        self.p = p
        self.interpolation = interpolation
        self.distortion_scale = distortion_scale

    def __call__(self, data):
        if isinstance(data, dict):
            height, width = F.get_image_size(data['image'])
            startpoints, endpoints = self.get_params(width, height, self.distortion_scale)
            for k,v in data.items():
                if v is not None:
                    data[k] = getattr(self, f'apply_{k}')(v, startpoints, endpoints)
        else:
            height, width = F.get_image_size(data)
            startpoints, endpoints = self.get_params(width, height, self.distortion_scale)
            data = self.apply_image(data, startpoints, endpoints)
        return data

    def apply_image(self, img, startpoints, endpoints):
        return F.perspective(img, startpoints, endpoints, self.interpolation)

    def apply_mask(self, mask, startpoints, endpoints):
        distort = partial(F.perspective, startpoints, endpoints, self.interpolation)
        if isinstance(mask, List):
            return [distort(m) for m in mask]
        return distort(mask)

    @staticmethod
    def get_params(width, height, distortion_scale):

        half_height = int(height / 2)
        half_width = int(width / 2)
        topleft = (random.randint(0, int(distortion_scale * half_width)),
                   random.randint(0, int(distortion_scale * half_height)))
        topright = (random.randint(width - int(distortion_scale * half_width) - 1, width - 1),
                    random.randint(0, int(distortion_scale * half_height)))
        botright = (random.randint(width - int(distortion_scale * half_width) - 1, width - 1),
                    random.randint(height - int(distortion_scale * half_height) - 1, height - 1))
        botleft = (random.randint(0, int(distortion_scale * half_width)),
                   random.randint(height - int(distortion_scale * half_height) - 1, height - 1))
        startpoints = [(0, 0), (width - 1, 0), (width - 1, height - 1), (0, height - 1)]
        endpoints = [topleft, topright, botright, botleft]
        return startpoints, endpoints

    def __repr__(self):
        return self.__class__.__name__

    def to_dict(self):
        return OrderedDict([('perspective', {
            'distortion_scale': self.distortion_scale,
            'p': self.p,
            'interpolation': self.interpolation
        })])



import cv2 as cv
import matplotlib.pyplot as plt
import numpy as np
import torch

import cvt

img = cv.imread('assets/image.jpg')

tfms = cvt.from_file('configs/transform1.json')
img = tfms(img)
plt.imshow(img)
plt.show()


import torch 

def vflip(img):
    return img.flip(-2)

def hflip(img):
    return img.flip(-1)

def normalize(tensor, mean, std):
    dtype = tensor.dtype
    mean = torch.as_tensor(mean, dtype=dtype, device=tensor.device)
    std = torch.as_tensor(std, dtype=dtype, device=tensor.device)
    tensor.sub_(mean[:, None, None]).div_(std[:, None, None])
    return tensor


import cv2 as cv
import numpy as np


def resize(img, size, interpolation=cv.INTER_LINEAR):
    '''
    Args:
        size: (h, w)
    '''

    if isinstance(size, int):
        h, w, _ = img.shape
        if w < h:
            ow = size
            oh = (ow / w) * h
        else:
            oh = size
            ow = (oh / h) * w
    else:
        oh, ow = size
    oh, ow = int(oh), int(ow)
    return cv.resize(img, (ow, oh), interpolation=interpolation)


def rotate(img, degree, interpolation=cv.INTER_LINEAR, expand=False, center=None):
    h, w, _ = img.shape
    if center is None:
        center = (w // 2, h // 2)
    M = cv.getRotationMatrix2D(center, degree, 1)

    if expand:
        cos, sin = abs(M[0, 0]), abs(M[0, 1])
        w, h = int(h * sin + w * cos), int(h * cos + w * sin)
        M[0, 2] += w/2 - center[0]
        M[1, 2] += h/2 - center[1]
    img = cv.warpAffine(img, M, (w, h), flags=interpolation)
    return img


# flipCode, horizontal: >0, vertical:0, horizaontal-vertical:<0
def hflip(img):
    return cv.flip(img, 1)


def vflip(img):
    return cv.flip(img, 0)


def crop(img, i, j, h, w):
    return img[i:(i+h), j:(j+w)]


def pad(img):
    pass


def center_crop(img, size):
    h, w, _ = img.shape
    if isinstance(size, int):
        size = (size, size)

    oh, ow = size
    i = int((h - oh) / 2)
    j = int((w - ow) / 2)
    return crop(img, i, j, oh, ow)


def resized_crop(img, i, j, h, w, size, interpolation=cv.INTER_LINEAR):
    img = crop(img, i, j, h, w)
    img = resize(img, size, interpolation)
    return img


def resized_center_crop(img, size, interpolation=cv.INTER_LINEAR):
    h, w, _ = img.shape
    img = center_crop(img, size)
    img = resize(img, (h, w), interpolation)
    return img


def adjust_brightness(img, scale):
    img = img * scale
    np.clip(img, 0, 255, out=img)
    return np.uint8(img)


def adjust_gamma(img, gamma, gain=1):
    gamma_map = np.array([255 * gain * pow(ele / 255., 1. / gamma)
                          for ele in range(256)]).astype(np.uint8)
    cv.LUT(img, gamma_map, img)
    return img

def to_gray(img, keep_dim=False):
    r, g, b = cv.split(img)
    gray = r * 299/1000 + g * 587/1000 + b * 114/1000
    if keep_dim:
        return cv.merge([gray] * 3)
    return gray

def adjust_saturation(img, scale):
    gray = to_gray(img, keep_dim=True)
    img = np.int32(img)
    img = img * scale + gray * (1.0 - scale) 
    np.clip(img, 0, 255, out=img)
    return np.uint8(img)

def adjust_contrast(img, scale):
    mean = round(to_gray(img).mean())
    gray = np.ones_like(img, dtype=np.uint8) * int(mean)
    img = img * scale + gray * (1.0 - scale) 
    np.clip(img, 0, 255, out=img)
    return np.uint8(img)

def adjust_hue(img, scale):

    if not(-0.5 <= scale <= 0.5):
        raise ValueError('hue_factor is not in [-0.5, 0.5].'.format(scale))
    img = cv.cvtColor(img, cv.COLOR_RGB2HSV)
    h, s, v = cv.split(img)

    h = np.int32(h)
    h = np.uint8((h + (scale * 180) % 181) % 181)

    img = cv.merge([h, s, v])
    img = cv.cvtColor(img, cv.COLOR_HSV2RGB)
    return img


def is_gray(img):
    if img.ndim == 2 or img.shape[-1] == 1:
        return True
    r, g, b = cv.split(img)
    if np.all(r == g) and np.all(r == b):
        return True
    return False


def clahe(img, clip_limit=2, tile_grid_size=(10, 10)):
    clahe = cv.createCLAHE(clip_limit, tile_grid_size)
    if is_gray(img):
        if img.ndim == 2 or img.shape[-1] == 1:
            img = clahe.apply(img)
        else:
            img = clahe.apply(img[..., 0])
            img = cv.cvtColor(img, cv.COLOR_GRAY2RGB)
    else:
        img = cv.cvtColor(img, cv.COLOR_RGB2LAB)
        img[..., 0] = clahe.apply(img[..., 0])
        img = cv.cvtColor(img, cv.COLOR_LAB2RGB)
    return img


def jpeg_quality(img, quality):
    _, im_code = cv.imencode('.jpg', img, (cv.IMWRITE_JPEG_QUALITY, quality))
    img = cv.imdecode(im_code, cv.IMREAD_UNCHANGED)
    return img


def perspective(img, startpoints, endpoints, interpolation=cv.INTER_LINEAR):
    M = cv.getPerspectiveTransform(startpoints, endpoints)
    img = cv.warpPerspective(img, M, img.shape[:2], interpolation)
    return img


def gaussian_noise(img, mean=0, var=0.01):
    noise = np.random.randn(*img.shape) * np.sqrt(var) + mean
    img = img * (1 + noise)
    np.clip(img, 0, 255, img)
    img = np.int8(img)
    return img

def poisson_noise(img, mean):
    noise = np.random.poisson(lam=mean, size=img.shape)
    np.sum(img, noise, out=img)
    np.clip(img, 0, 255, out=img)
    return np.uint8(img)


def speckle_noise(img):
    low = mean - np.sqrt(var)
    high = mean+ np.sqrt(var)
    J = np.random.uniform(low=low, high=high, size=size)
    return img

def imnoise(img, noise_type, mean=None, var=None, local_var=None, d=None):
    assert noise_type in ['gaussian', 'localvar', 'poisson', 'speckle']
    if noise_type == 'gaussian':
        img = gaussian_noise(img)
    elif noise_type == 'poisson':
        img = poisson_noise(img)
    elif  noise_type == 'speckle':
        img = speckle_noise(img)


    img  = img * (1 + J) * 255
    np.clip(img, 0, 255, out=img)
    return np.uint8(img)


def label_map(img, table, value=0):
    table = np.uint8([table.get(x, value) for x in range(256)])
    cv.LUT(img, table, img)
    return img


from setuptools import find_packages, setup

install_requires=[
    'torch',
    'opencv-python>=3.4.3.18',
    'numpy',
    'pillow'
    ]

setup(
    name='cvt',
    version='0.0.9',
    packages=find_packages(exclude=('test',)),
    author='jie.wang',
    author_email='jie.wang@ruc.edu.cn',
    description='Image Transformation Package',
    url='https://github.com/wangjie-ruc/cvt',
    license='BSD',
    install_requires=install_requires
)
